@@ -97,7 +97,7 @@ G_DEFINE_QUARK (autoar-extractor, autoar_extractor)
 #define BUFFER_SIZE (64 * 1024)
 #define NOT_AN_ARCHIVE_ERRNO 2013
 #define EMPTY_ARCHIVE_ERRNO 2014
-#define WRONG_PASSPHRASE_ERRNO 2015
+#define INCORRECT_PASSPHRASE_ERRNO 2015
 
 typedef struct _GFileAndInfo GFileAndInfo;
 
@@ -134,8 +134,6 @@ struct _AutoarExtractor
 
   GList *files_list;
 
-  gchar *passphrase;
-
   GHashTable *userhash;
   GHashTable *grouphash;
   GArray     *extracted_dir_list;
@@ -148,6 +146,9 @@ struct _AutoarExtractor
 
   int in_thread         : 1;
   int use_raw_format    : 1;
+
+  gchar *passphrase;
+  gboolean passphrase_requested;
 };
 
 G_DEFINE_TYPE (AutoarExtractor, autoar_extractor, G_TYPE_OBJECT)
@@ -163,10 +164,10 @@ enum
   SCANNED,
   DECIDE_DESTINATION,
   PROGRESS,
-  REQUEST_PASSPHRASE,
   CONFLICT,
   CANCELLED,
   COMPLETED,
+  REQUEST_PASSPHRASE,
   AR_ERROR,
   LAST_SIGNAL
 };
@@ -515,6 +516,10 @@ autoar_extractor_dispose (GObject *object)
     self->extracted_dir_list = NULL;
   }
 
+  if (self->passphrase != NULL) {
+    g_free (self->passphrase);
+  }
+
   G_OBJECT_CLASS (autoar_extractor_parent_class)->dispose (object);
 }
 
@@ -688,37 +693,16 @@ libarchive_read_skip_cb (struct archive *ar_read,
   return 0;
 }
 
-static inline void
-autoar_extractor_signal_request_passphrase (AutoarExtractor *self)
+static inline gchar *
+autoar_extractor_request_passphrase (AutoarExtractor *self)
 {
-  autoar_common_g_signal_emit (self, self->in_thread,
-                               autoar_extractor_signals[REQUEST_PASSPHRASE], 0,
-                               &self->passphrase);
-}
-
-static const char *
-libarchive_read_set_passphrase_cb (struct archive *ar_read,
-                                   void           *client_data)
-{
-  AutoarExtractor *self = AUTOAR_EXTRACTOR (client_data);
-  GError *error = NULL;
-
-  if (self->passphrase == NULL) {
-    autoar_extractor_signal_request_passphrase (self);
-
-    return self->passphrase;
+  if (!self->passphrase_requested) {
+    autoar_common_g_signal_emit (self, self->in_thread,
+                                 autoar_extractor_signals[REQUEST_PASSPHRASE], 0, &self->passphrase);
+    self->passphrase_requested = TRUE;
   }
 
-  error = g_error_new (AUTOAR_EXTRACTOR_ERROR,
-                       WRONG_PASSPHRASE_ERRNO,
-                       "\'%s\': %s",
-                       self->source_basename,
-                       "wrong passphrase");
-  autoar_common_g_signal_emit (self, self->in_thread,
-                               autoar_extractor_signals[AR_ERROR], 0,
-                               error);
-
-  return NULL;
+  return self->passphrase;
 }
 
 static int
@@ -738,7 +722,10 @@ libarchive_create_read_object (gboolean          use_raw_format,
   archive_read_set_seek_callback (*a, libarchive_read_seek_cb);
   archive_read_set_skip_callback (*a, libarchive_read_skip_cb);
   archive_read_set_callback_data (*a, self);
-  archive_read_set_passphrase_callback (*a, self, libarchive_read_set_passphrase_cb);
+
+  if (self->passphrase != NULL) {
+    archive_read_add_passphrase (*a, self->passphrase);
+  }
 
   return archive_read_open1 (*a);
 }
@@ -881,32 +868,70 @@ autoar_extractor_get_common_prefix (GList *files,
   return prefix;
 }
 
+static gboolean
+is_valid_filename (GFile *file, GFile *destination)
+{
+  g_autoptr (GFile) parent = NULL;
+  g_autoptr (GFileInfo) info = NULL;
+  g_autoptr (GError) error = NULL;
+
+  if (g_file_equal (file, destination))
+    return TRUE;
+
+  if (!g_file_has_prefix (file, destination))
+    return FALSE;
+
+  /* Resolve symbolic link ancestors to confirm file is actually inside destination. */
+  parent = g_file_get_parent (file);
+  info = g_file_query_info (parent,
+                            G_FILE_ATTRIBUTE_STANDARD_IS_SYMLINK ","
+                            G_FILE_ATTRIBUTE_STANDARD_SYMLINK_TARGET,
+                            G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
+                            NULL,
+                            &error);
+
+  /* The parent directories don't have to be created yet. */
+  if (error && !g_error_matches (error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND))
+    return FALSE;
+
+  if (info && g_file_info_get_is_symlink (info)) {
+    g_autoptr (GFile) cwd = NULL;
+    const gchar *target;
+
+    target = g_file_info_get_symlink_target (info);
+    if (g_path_is_absolute (target))
+      return FALSE;
+
+    cwd = g_file_get_parent (parent);
+    g_object_unref (parent);
+    parent = g_file_resolve_relative_path (cwd, target);
+  }
+
+  /* Climb up the path to resolve every symbolic link ancestor found */
+  return is_valid_filename (parent, destination);
+}
+
 static GFile*
 autoar_extractor_do_sanitize_pathname (AutoarExtractor *self,
                                        const char      *pathname_bytes)
 {
   GFile *extracted_filename;
   gboolean valid_filename;
-  g_autofree char *sanitized_pathname;
+  g_autofree char *sanitized_pathname = NULL;
   g_autofree char *utf8_pathname;
 
   utf8_pathname = autoar_common_get_utf8_pathname (pathname_bytes);
   extracted_filename = g_file_get_child (self->destination_dir,
                                          utf8_pathname ?  utf8_pathname : pathname_bytes);
 
-  valid_filename =
-    g_file_equal (extracted_filename, self->destination_dir) ||
-    g_file_has_prefix (extracted_filename, self->destination_dir);
-
+  valid_filename = is_valid_filename (extracted_filename, self->destination_dir);
   if (!valid_filename) {
-    g_autofree char *basename;
-
-    basename = g_file_get_basename (extracted_filename);
+    g_warning ("autoar_extractor_do_sanitize_pathname: %s is outside of the destination dir",
+                g_file_peek_path (extracted_filename));
 
     g_object_unref (extracted_filename);
 
-    extracted_filename = g_file_get_child (self->destination_dir,
-                                           basename);
+    return NULL;
   }
 
   if (self->prefix != NULL && self->new_prefix != NULL) {
@@ -967,6 +992,57 @@ autoar_extractor_check_file_conflict (GFile  *file,
   return conflict;
 }
 
+static gboolean
+autoar_extractor_delete_file_recursively (AutoarExtractor *self,
+                                          GFile *file)
+{
+  gboolean success;
+  g_autoptr (GError) error = NULL;
+
+  do {
+    g_autoptr (GFileEnumerator) enumerator = NULL;
+
+    success = g_file_delete (file, self->cancellable, &error);
+    if (success || !g_error_matches (error, G_IO_ERROR, G_IO_ERROR_NOT_EMPTY)) {
+      break;
+    }
+
+    g_clear_error (&error);
+    enumerator = g_file_enumerate_children (file,
+                                            G_FILE_ATTRIBUTE_STANDARD_NAME,
+                                            G_FILE_QUERY_INFO_NONE,
+                                            self->cancellable, &error);
+    if (enumerator) {
+      GFileInfo *info;
+
+      success = TRUE;
+
+      info = g_file_enumerator_next_file (enumerator,
+                                          self->cancellable,
+                                          &error);
+      while (info != NULL) {
+        g_autoptr (GFile) child = NULL;
+
+        child = g_file_enumerator_get_child (enumerator, info);
+
+        success = success && autoar_extractor_delete_file_recursively (self, child);
+        g_object_unref (info);
+
+        info = g_file_enumerator_next_file (enumerator,
+                                            self->cancellable,
+                                            &error);
+      }
+    }
+
+    if (error != NULL) {
+      success = FALSE;
+    }
+
+  } while (success);
+
+  return success;
+}
+
 static void
 autoar_extractor_do_write_entry (AutoarExtractor      *self,
                                  struct archive       *a,
@@ -976,7 +1052,9 @@ autoar_extractor_do_write_entry (AutoarExtractor      *self,
 {
   GFileInfo *info;
   mode_t filetype;
+#if defined HAVE_LINK || defined HAVE_MKNOD || defined HAVE_MKFIFO
   int r;
+#endif
 
   {
     GFile *parent;
@@ -1028,10 +1106,10 @@ autoar_extractor_do_write_entry (AutoarExtractor      *self,
   /* user */
   {
     guint32 uid;
-    const char *uname;
 
     g_debug ("autoar_extractor_do_write_entry: user");
 #ifdef HAVE_GETPWNAM
+    const char *uname;
     if ((uname = archive_entry_uname (entry)) != NULL) {
       void *got_uid;
       if (g_hash_table_lookup_extended (self->userhash, uname, NULL, &got_uid) == TRUE) {
@@ -1057,10 +1135,10 @@ autoar_extractor_do_write_entry (AutoarExtractor      *self,
   /* group */
   {
     guint32 gid;
-    const char *gname;
 
     g_debug ("autoar_extractor_do_write_entry: group");
 #ifdef HAVE_GETGRNAM
+    const char *gname;
     if ((gname = archive_entry_gname (entry)) != NULL) {
       void *got_gid;
       if (g_hash_table_lookup_extended (self->grouphash, gname, NULL, &got_gid) == TRUE) {
@@ -1106,7 +1184,9 @@ autoar_extractor_do_write_entry (AutoarExtractor      *self,
 #endif
 
   g_debug ("autoar_extractor_do_write_entry: writing");
+#if defined HAVE_MKNOD || defined HAVE_MKFIFO
   r = 0;
+#endif
 
   switch (filetype = archive_entry_filetype (entry)) {
     default:
@@ -1133,7 +1213,7 @@ autoar_extractor_do_write_entry (AutoarExtractor      *self,
         if (ostream != NULL) {
           /* Archive entry size may be zero if we use raw format. */
           if (archive_entry_size(entry) > 0 || self->use_raw_format) {
-            while (archive_read_data_block (a, &buffer, &size, &offset) == ARCHIVE_OK) {
+            while ((r = archive_read_data_block (a, &buffer, &size, &offset)) == ARCHIVE_OK) {
               /* buffer == NULL occurs in some zip archives when an entry is
                * completely read. We just skip this situation to prevent GIO
                * warnings. */
@@ -1160,6 +1240,20 @@ autoar_extractor_do_write_entry (AutoarExtractor      *self,
               self->completed_size += written;
               autoar_extractor_signal_progress (self);
             }
+
+            if (r == ARCHIVE_FAILED) {
+              autoar_extractor_delete_file_recursively (self, self->destination_dir);
+              if (self->error == NULL) {
+                self->error = g_error_new (AUTOAR_EXTRACTOR_ERROR,
+                                           INCORRECT_PASSPHRASE_ERRNO,
+                                           "%s",
+                                           archive_error_string (a));
+              }
+              g_output_stream_close (ostream, self->cancellable, NULL);
+              g_object_unref (ostream);
+
+              return;
+            }
           }
           g_output_stream_close (ostream, self->cancellable, NULL);
           g_object_unref (ostream);
@@ -1272,7 +1366,9 @@ autoar_extractor_do_write_entry (AutoarExtractor      *self,
   }
 #endif
 
+#ifdef HAVE_LINK
 applyinfo:
+#endif
   g_debug ("autoar_extractor_do_write_entry: applying info");
   g_file_set_attributes_from_info (dest,
                                    info,
@@ -1443,15 +1539,6 @@ autoar_extractor_class_init (AutoarExtractorClass *klass)
                   G_TYPE_UINT64,
                   G_TYPE_UINT);
 
-
-  autoar_extractor_signals[REQUEST_PASSPHRASE] =
-    g_signal_new ("request-passphrase",
-                  type,
-                  G_SIGNAL_RUN_LAST,
-                  0, NULL, NULL,
-                  g_cclosure_marshal_generic,
-                  G_TYPE_POINTER,
-                  0);
 /**
  * AutoarExtractor::conflict:
  * @self: the #AutoarExtractor
@@ -1506,6 +1593,22 @@ autoar_extractor_class_init (AutoarExtractorClass *klass)
                   G_TYPE_NONE,
                   0);
 
+/**
+ * AutoarExtractor::request-passphrase:
+ * @self: the #AutoarExtractor
+ *
+ * This signal is emitted when the archive extracting job needs a
+ * passphrase.
+ **/
+  autoar_extractor_signals[REQUEST_PASSPHRASE] =
+    g_signal_new ("request-passphrase",
+                  type,
+                  G_SIGNAL_RUN_LAST,
+                  0, NULL, NULL,
+                  g_cclosure_marshal_generic,
+                  G_TYPE_STRING,
+                  0);
+
 /**
  * AutoarExtractor::error:
  * @self: the #AutoarExtractor
@@ -1559,6 +1662,9 @@ autoar_extractor_init (AutoarExtractor *self)
 
   self->in_thread = FALSE;
   self->use_raw_format = FALSE;
+
+  self->passphrase = NULL;
+  self->passphrase_requested = FALSE;
 }
 
 /**
@@ -1639,6 +1745,10 @@ autoar_extractor_step_scan_toplevel (AutoarExtractor *self)
       return;
     }
 
+    if (archive_entry_is_encrypted (entry)) {
+      autoar_extractor_request_passphrase (self);
+    }
+
     if (self->use_raw_format) {
       pathname = autoar_common_get_basename_remove_extension (g_file_get_path(self->source_file));
       g_debug ("autoar_extractor_step_scan_toplevel: %d: raw pathname = %s",
@@ -1862,10 +1972,18 @@ autoar_extractor_step_extract (AutoarExtractor *self) {
 
     extracted_filename =
       autoar_extractor_do_sanitize_pathname (self, pathname);
+    if (extracted_filename == NULL) {
+      archive_read_data_skip (a);
+      continue;
+    }
 
     if (hardlink != NULL) {
       hardlink_filename =
         autoar_extractor_do_sanitize_pathname (self, hardlink);
+        if (hardlink_filename == NULL) {
+          archive_read_data_skip (a);
+          continue;
+        }
     }
 
     /* Attempt to solve any name conflict before doing any operations */
