@@ -21,7 +21,7 @@
  *
  */
 
- #include "autoar-mime-types.h"
+#include "autoar-mime-types.h"
 
 static gchar *supported_mime_types[] = {
   "application/x-7z-compressed",
@@ -31,7 +31,6 @@ static gchar *supported_mime_types[] = {
   "application/x-compress",
   "application/x-compressed-tar",
   "application/x-cpio",
-  "application/x-gzip",
   "application/x-lha",
   "application/x-lzip",
   "application/x-lzip-compressed-tar",
@@ -48,22 +47,6 @@ static gchar *supported_mime_types[] = {
   NULL
 };
 
-static GHashTable *supported_mime_types_table = NULL;
-
-static void
-initialize_supported_mime_types_table (void)
-{
-  int i;
-
-  supported_mime_types_table = g_hash_table_new (g_str_hash,
-                                                 g_str_equal);
-
-  for (i = 0; supported_mime_types[i] != NULL; ++i) {
-    g_hash_table_add (supported_mime_types_table,
-                      supported_mime_types[i]);
-  }
-}
-
 /**
  * autoar_check_mime_type_supported:
  * @mime_type: a string representing the mime type
@@ -71,16 +54,18 @@ initialize_supported_mime_types_table (void)
  * Checks whether a mime type is supported by autoar. This function does no
  * blocking IO.
  *
- * Returns: an #AutoarFilter
+ * Returns: %TRUE if the mime type is supported
  **/
 gboolean
 autoar_check_mime_type_supported (const gchar *mime_type)
 {
-  if (supported_mime_types_table == NULL) {
-    initialize_supported_mime_types_table ();
+  gint i;
+  for (i = 0; supported_mime_types[i] != NULL; ++i) {
+    if (g_content_type_equals (supported_mime_types[i], mime_type))
+      return TRUE;
   }
 
-  return g_hash_table_contains (supported_mime_types_table, mime_type);
+  return FALSE;
 }
 
 /**
@@ -90,7 +75,7 @@ autoar_check_mime_type_supported (const gchar *mime_type)
  * This function will query the file's mime type and then call
  * autoar_check_mime_type_supported(), so it does blocking IO.
  *
- * Returns: an #AutoarFilter
+ * Returns: %TRUE if the mime type of the #GFile is supported
  **/
 gboolean
 autoar_query_mime_type_supported (GFile *file)
